from past.types import basestring
import warnings

from . import util


class BaseCurve:
    def __init__(self, id, metadata, session):
        self._metadata = metadata
        self._session = session
        self.time_zone = 'CET'
        if metadata is None:
            self.hasMetadata = False
        else:
            self.hasMetadata = True
            for key, val in metadata.items():
                setattr(self, key, val)
        self.id = id
        self.tz = util.parse_tz(self.time_zone)

    def __str__(self):
        if hasattr(self, 'curve_type'):
            curve_type = self.curve_type
        else:
            curve_type = 'UNKNOWN'
        if hasattr(self, 'name'):
            name = self.name
        else:
            name = str(self.id)
        return "{}({})".format(curve_type, name)

    def _add_from_to(self, args, first, last, prefix=''):
        if first is not None:
            args.append(util.make_arg('{}from'.format(prefix), first))
        if last is not None:
            args.append(util.make_arg('{}to'.format(prefix), last))

    def _add_functions(self, args, time_zone, filter, function, frequency, output_time_zone):
        if time_zone is not None:
            args.append(util.make_arg('time_zone', time_zone))
        if filter is not None:
            args.append(util.make_arg('filter', filter))
        if function is not None:
            args.append(util.make_arg('function', function))
        if frequency is not None:
            args.append(util.make_arg('frequency', frequency))
        if output_time_zone is not None:
            args.append(util.make_arg('output_time_zone', output_time_zone))

    def _load_data(self, url, failmsg, urlbase=None):
        if urlbase is None:
            urlbase = self._session.urlbase
        response = self._session.data_request('GET', urlbase, url)
        self._last_response = response
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 204 or response.status_code == 404:
            return None
        raise util.CurveException('{}: {} ({})'.format(failmsg, response.content, response.status_code))

    def access(self):
        url = '/api/curves/{}/access'.format(self.id)
        return self._load_data(url, 'Failed to load curve access')


class TimeSeriesCurve(BaseCurve):
    def get_data(self, data_from=None, data_to=None, time_zone=None, filter=None,
                 function=None, frequency=None, output_time_zone=None):
        """ Getting data from Time Series curves

        A Time Series curves holds a single time series.
        This is used for actual values, backcasts, normals, etc. This function
        fetches data between two given timestamps. It also possible to process
        the curve directly in the API using filter and aggregation functions.
        This can be used with great effect to reduce the amount of data
        retrieved if the full set of details is not needed.
        All time series are returned in a :class:`wapi.util.TS` object.

        Parameters
        ----------

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in any of the following types (also valid without time):

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as ``data_from``.
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters`.

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions`.

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function. You can find the valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """

        args = []
        astr = ''
        self._add_from_to(args, data_from, data_to)
        self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if len(args) > 0:
            astr = '?{}'.format('&'.join(args))
        url = '/api/series/{}{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load curve data')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.TIME_SERIES)


class TaggedCurve(BaseCurve):
    def get_tags(self):
        """ Get list of available tags for this curve

        Returns
        -------
        list
            Returns a list of all available tags for a Tagged Instance curve.
        """
        url = '/api/series/tagged/{}/tags'.format(self.id)
        return self._load_data(url, 'Failed to fetch tags')

    def get_data(self, tag=None, data_from=None, data_to=None, time_zone=None, filter=None,
                 function=None, frequency=None, output_time_zone=None):
        """ Getting data from TAGGED curves

        A tagged curve holds a set of closely related time series, each
        identified by a tag. The most common use of tags is for ensemble
        weather data.

        This function fetches data for one or multiple given tags and returns
        a list of :class:`wapi.util.TS` objects.
        It also possible to process the curve directly in the API using filter
        and aggregation functions. This can be used with great effect to reduce
        the amount of data retrieved if the full set of details is not needed.
        All time series are returned in a :class:`wapi.util.TS` object.

        Parameters
        ----------

        tag: str or list, optional
            tag or tags to get the data for. If omitted, the default
            tag is returned. If a list of multiple tags is given, the function
            will return a list with a :class:`wapi.util.TS` object for each tag.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in any of the following types (also valid without time):

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as ``data_from``.
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        unwrap = False
        if tag is None:
            args = []
            unwrap = True
        else:
            if isinstance(tag, basestring):
                unwrap = True
            args=[util.make_arg('tag', tag)]
        self._add_from_to(args, data_from, data_to)
        self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        astr = '&'.join(args)
        url = '/api/series/tagged/{}?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load tagged curve data')
        if result is None:
            return result
        res = [util.TS(input_dict=r, curve_type=util.TAGGED) for r in result]
        if unwrap and len(res) == 1:
            res = res[0]
        return res


class InstanceCurve(BaseCurve):
    def search_instances(self, issue_date_from=None, issue_date_to=None,
                         issue_dates=None, issue_weekdays=None, issue_days=None, issue_months=None,
                         issue_times=None, with_data=False, data_from=None, data_to=None,
                         time_zone=None, filter=None, function=None, frequency=None,
                         output_time_zone=None, only_accessible=None, modified_since=None):
        """ Getting data from INSTANCE curves for multiple issue_dates

        An INSTANCE curve typically represents forecast,
        and contains a time series for each issue_date of the forecast.
        This function returns a list of time series for all available
        issue_dates (within a given period, if specified)
        as a a list of :class:`wapi.util.TS` objects.
        It also possible to process
        the curve directly in the API using filter and aggregation functions.
        This can be used with great effect to reduce the amount of data
        retrieved if the full set of details is not needed.
        By default this function returns the :class:`wapi.util.TS` objects
        without data, which can be change by setting the "with_data" argument
        to True.

        Parameters
        ----------

        issue_date_from: time-stamp, optional
            Limits the timerange to return all available issue_dates.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange to return for the latest available issue_date.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps to return :class:`wapi.util.TS` objects for.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        issue_weekdays: list of str, optional
            Filter issue_date on day of week.

        issue_days: list of int, optional
            Filter issue_date on day of month

        issue_months: list of str, optional
            Filter issue_date on month

        issue_times: list of str, optional
            Filter issue_date on time of day
            Format is 'HH', 'HH:mm' or 'HH:mm:ss'.

        with_data: bool, optional
            If with_data is False, the returned  :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        modified_since: datestring, pandas.Timestamp or datetime.datetime
            only return instances that where modified after given datetime.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower())]
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        if with_data:
            self._add_from_to(args, data_from, data_to, prefix='data_')
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        if issue_weekdays is not None:
            args.append(util.make_arg('issue_weekday', issue_weekdays))
        if issue_days is not None:
            args.append(util.make_arg('issue_day', issue_days))
        if issue_months is not None:
            args.append(util.make_arg('issue_month', issue_months))
        if issue_times is not None:
            args.append(util.make_arg('issue_time', issue_times))
        if modified_since is not None:
            args.append(util.make_arg('modified_since', modified_since))
        astr = '&'.join(args)
        url = '/api/instances/{}?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find instances')
        if result is None:
            return result
        return [util.TS(input_dict=r, curve_type=util.INSTANCES) for r in result]

    def get_instance(self, issue_date, with_data=True, data_from=None, data_to=None,
                     time_zone=None, filter=None, function=None, frequency=None,
                     output_time_zone=None, only_accessible=None):
        """ Getting data from INSTANCE curves for a specific issue_date

        An INSTANCE curve typically represents forecast,
        and contains a time series for each issue_date of the forecast.
        This function returns the time series for a specified issue_date
        as a :class:`wapi.util.TS` object. It also possible to process
        the curve directly in the API using filter and aggregation functions.
        This can be used with great effect to reduce the amount of data
        retrieved if the full set of details is not needed.

        Parameters
        ----------
        issue_date: time-stamp
            Time-stamp representing the issue date to get data for.
            The timestamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        with_data: bool, optional
            If with_data is False, the returned  :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower()),
              util.make_arg('issue_date', issue_date)]
        if with_data:
            self._add_from_to(args, data_from, data_to)
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        astr = '&'.join(args)
        url = '/api/instances/{}/get?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load instance')
        if result is None:
            return result
        return util.TS(input_dict=result, issue_date=issue_date, curve_type=util.INSTANCES)

    def get_latest(self, issue_date_from=None, issue_date_to=None, issue_dates=None,
                   with_data=True, data_from=None, data_to=None, time_zone=None, filter=None,
                   function=None, frequency=None, output_time_zone=None, only_accessible=None):
        """ Getting data from INSTANCE curves for the latest available issue_date

        An INSTANCE curve typically represents forecast,
        and contains a time series for each issue_date of the forecast.
        This function returns the time series for the latest available
        issue_date (within a given period, if specified)
        as a :class:`wapi.util.TS` object. It also possible to process
        the curve directly in the API using filter and aggregation functions.
        This can be used with great effect to reduce the amount of data
        retrieved if the full set of details is not needed.

        Parameters
        ----------

        issue_date_from: time-stamp, optional
            Limits the timerange to search for the latest available issue_date.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange to search for the latest available issue_date.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps in which to search for the latest available
            issue_date.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        with_data: bool, optional
            If with_data is False, the returned  :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower())]
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        if with_data:
            self._add_from_to(args, data_from, data_to, prefix='data_')
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        astr = '&'.join(args)
        url = '/api/instances/{}/latest?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load instance')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.INSTANCES)

    def get_relative(self, data_offset, data_max_length=None, issue_date_from=None, issue_date_to=None,
                     issue_dates=None, issue_weekdays=None, issue_days=None, issue_months=None, issue_times=None,
                     data_from=None, data_to=None, time_zone=None, filter=None, function=None,
                     frequency=None, output_time_zone=None):
        """ Get a relative forecast from the INSTANCE curve

        A relative forecast is a time series created by joining multiple
        instances so that the data date is issue_date + data_offset.
        If the data frequency is higher than the frequency of the selected
        instances, a range of data values from each instance will be used.
        Similarly if the issue frequency is higher than the data frequency,
        the same data date will be used from several instances.

        Parameters allow control of the set of instances used and how they
        are joined together, as well as post-processing of the result.

        Parameters
        ----------

        data_offset: duration, mandatory
            The duration added to the issue_date to find the start of the data
            fragment.  Format is an ISO-8601 duration string up to "days".

        data_max_length: duration, optional
            The longest duration selected from a single instance, mostly used
            to detect missing instances.  ISO-8601 duration string up to "days".

        issue_date_from: time-stamp, optional
            Limits the timerange used to select instances.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps used to select instances.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        issue_weekdays: list of strings, optional
            Limits the instances to those matching the given weekdays.

        issue_days: list of integers, optional
            Limits the instances to those matching the given days of month.

        issue_months: list of strings, optional
            Limits the instances to those matching the given months.

        issue_times: list of strings, optional
            Limits the instances to those matching the given times of day.
            Format is 'HH', 'HH:mm' or 'HH:mm:ss'.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be determined by the first
            selected instance and the data_offset parameter. If only the date
            (without time) is given, the time is assumed to be 00:00. The
            timestamp can be provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available in the last instance selected,
            optionally limited by the data_max_length parameter.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        args = [util.make_arg('data_offset', '{}'.format(data_offset))]
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        self._add_from_to(args, data_from, data_to, prefix='data_')
        self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if data_max_length is not None:
            args.append(util.make_arg('data_max_length', data_max_length))
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        if issue_weekdays is not None:
            args.append(util.make_arg('issue_weekday', issue_weekdays))
        if issue_days is not None:
            args.append(util.make_arg('issue_day', issue_days))
        if issue_months is not None:
            args.append(util.make_arg('issue_month', issue_months))
        if issue_times is not None:
            args.append(util.make_arg('issue_time', issue_times))
        astr = '&'.join(args)
        url = '/api/instances/{}/relative?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find instances')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.INSTANCES)

    def get_absolute(self, data_date, issue_frequency=None, issue_date_from=None, issue_date_to=None):
        """ Get an absolute forecast from the INSTANCE curve

        An absolute forecast is a time series created by selecting a single
        data date from a range of instances, using the issue_date as
        data date in the result.

        Parameters
        ----------

        data_date: time-stamp, mandatory
            The data date of the absolute forecast.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_frequency: str, optional
            The frequency of the returned time series.  Mandatory if the
            issue frequency of the curve is 'ANY'.  Data will only be
            returned for those instances whose issue_date is compatible
            with issue_frequency.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        issue_date_from: time-stamp, optional
            The start of the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "data_date".

        issue_date_to: time-stamp, optional
            The end of the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "data_date".

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        args = [util.make_arg('data_date', data_date)]
        if issue_frequency is not None:
            args.append(util.make_arg('issue_frequency', issue_frequency))
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        astr = '&'.join(args)
        url = '/api/instances/{}/absolute?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find instances')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.INSTANCES)


class TaggedInstanceCurve(BaseCurve):
    def get_tags(self):
        """ Get list of available tags for this curve

        Returns
        -------
        list
            Returns a list of all available tags for a Tagged Instance curve.
        """
        url = '/api/instances/tagged/{}/tags'.format(self.id)
        return self._load_data(url, 'Failed to fetch tags')

    def search_instances(self, tags=None, issue_date_from=None, issue_date_to=None,
                         issue_dates=None, issue_weekdays=None, issue_days=None, issue_months=None,
                         issue_times=None, with_data=False, data_from=None, data_to=None,
                         time_zone=None, filter=None, function=None, frequency=None,
                         output_time_zone=None, only_accessible=None, modified_since=None):
        """ Getting data from TAGGED_INSTANCE curves for multiple issue_dates

        A TAGGED INSTANCE curve typically represents forecast that contain
        multiple time series for each issue_date of the forecast, which are
        assigned to so called tags. Each timeseries is therefore defined by a
        unique combination of issue_date and tag. Ensamble forecasts are a
        typical use case for TAGGED INSTANCE curves.

        This function returns a list of time series for all available
        issue_dates (within a given period, if specified) and tags (from a
        given list of tags, if specified) as a a list of :class:`wapi.util.TS`
        objects.

        It also possible to process the curve directly in the API using filter
        and aggregation functions. This can be used with great effect to reduce
        the amount of data retrieved if the full set of details is not needed.
        By default this function returns the :class:`wapi.util.TS` objects
        without data, which can be change by setting the "with_data" argument
        to True.

        Parameters
        ----------

        tags: str or list, optional
            tag or tags to consider. The function will only return objects with
            tags defined here. If None (default) is given, all available tags
            are considered.

        issue_date_from: time-stamp, optional
            Limits the timerange to return all available issue_dates.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange to return for the latest available issue_date.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps to return :class:`wapi.util.TS` objects for.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        issue_weekdays: list of str, optional
            Filter issue_date on day of week.

        issue_days: list of int, optional
            Filter issue_date on day of month

        issue_months: list of str, optional
            Filter issue_date on month

        issue_times: list of str, optional
            Filter issue_date on time of day
            Format is 'HH', 'HH:mm' or 'HH:mm:ss'.

        with_data: bool, optional
            If with_data is False, the returned :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        modified_since: datestring, pandas.Timestamp or datetime.datetime
            only return instances that where modified after given datetime.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower())]
        if tags is not None:
            args.append(util.make_arg('tag', tags))
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        if with_data:
            self._add_from_to(args, data_from, data_to, prefix='data_')
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        if issue_weekdays is not None:
            args.append(util.make_arg('issue_weekday', issue_weekdays))
        if issue_days is not None:
            args.append(util.make_arg('issue_day', issue_days))
        if issue_months is not None:
            args.append(util.make_arg('issue_month', issue_months))
        if issue_times is not None:
            args.append(util.make_arg('issue_time', issue_times))
        if modified_since is not None:
            args.append(util.make_arg('modified_since', modified_since))
        astr = '&'.join(args)
        url = '/api/instances/tagged/{}?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find tagged instances')
        if result is None:
            return result
        return [util.TS(input_dict=r, curve_type=util.TAGGED_INSTANCES) for r in result]

    def get_instance(self, issue_date, tag=None, with_data=True, data_from=None, data_to=None,
                     time_zone=None, filter=None, function=None, frequency=None,
                     output_time_zone=None, only_accessible=None):
        """ Getting data from TAGGED_INSTANCE curves for a specific issue_date

        A TAGGED INSTANCE curve typically represents forecast that contain
        multiple time series for each issue_date of the forecast, which are
        assigned to so called tags. Each timeseries is therefore defined by a
        unique combination of issue_date and tag. Ensamble forecasts are a
        typical use case for TAGGED INSTANCE curves.

        This function returns a time series for the combinations of a specified
        issue_date and all given tags as a lit of :class:`wapi.util.TS` objects.
        It also possible to process the curve directly in the API using filter
        and aggregation functions. This can be used with great effect to reduce
        the amount of data retrieved if the full set of details is not needed.

        Parameters
        ----------

        issue_date: time-stamp
            Time-stamp representing the issue date to get data for.
            The timestamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        tag: str or list, optional
            tag or tags to get the data for. If omitted, the default
            tag is returned. If a list of multiple tags is given, the function
            will return a list with a :class:`wapi.util.TS` object for each tag.

        with_data: bool, optional
            If with_data is False, the returned  :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """

        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower()),
              util.make_arg('issue_date', issue_date)]
        unwrap = False
        if tag is None:
            unwrap = True
        else:
            if isinstance(tag, basestring):
                unwrap = True
            args.append(util.make_arg('tag', tag))
        if with_data:
            self._add_from_to(args, data_from, data_to)
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        astr = '&'.join(args)
        url = '/api/instances/tagged/{}/get?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load tagged instance')
        if result is None:
            return result
        res = [util.TS(input_dict=r, issue_date=issue_date, curve_type=util.TAGGED_INSTANCES) for r in result]
        if unwrap and len(res) == 1:
            res = res[0]
        return res

    def get_latest(self, tags=None, issue_date_from=None, issue_date_to=None, issue_dates=None,
                   with_data=True, data_from=None, data_to=None, time_zone=None, filter=None,
                   function=None, frequency=None, output_time_zone=None, only_accessible=None):
        """ Getting data from TAGGED INSTANCE curves for the latest issue_date

        A TAGGED INSTANCE curve typically represents forecasts that contain
        multiple time series for each issue_date of the forecast, which are
        assigned to so called tags. Each timeseries is therefore defined by a
        unique combination of issue_date and tag. Ensamble forecasts are a
        typical use case for TAGGED INSTANCE curves.

        This function returns the time series for ONE tag and the latest
        available issue_date (within a given period, if specified)
        as :class:`wapi.util.TS` objects. The tag to get the data for
        can be specified in the "tags" argument. If None (=all tags)
        or a list of tags is provided, only the timeseries for one of the tags
        (the first one found to have valid data) is returned. So it is
        recommended to specify ONE desired tag in "tags".

        It also possible to process
        the curve directly in the API using filter and aggregation functions.
        This can be used with great effect to reduce the amount of data
        retrieved if the full set of details is not needed.

        Parameters
        ----------
        tags: string or list, optional
            tag or list of tags to consider when returning the
            :class:`wapi.util.TS` object. The function returns a timeseries
            for ONE of the given tags. If you want get data for a specific tag,
            only specify one tag here (recommended!).

        issue_date_from: time-stamp, optional
            Limits the timerange to search for the latest available issue_date.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange to search for the latest available issue_date.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps in which to search for the latest available
            issue_date.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        with_data: bool, optional
            If with_data is False, the returned  :class:`wapi.util.TS` object
            only contains the attributes and meta data information but no
            data values.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be the first date with data
            available. If only the date (without time) is
            given, the time is assumed to be 00:00. The timestamp can be
            provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """

        if only_accessible is not None:
            warnings.warn("only_accessible parameter will be removed soon.", FutureWarning, stacklevel=2)
        args=[util.make_arg('with_data', '{}'.format(with_data).lower())]
        if tags is not None:
            args.append(util.make_arg('tag', tags))
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        if with_data:
            self._add_from_to(args, data_from, data_to, prefix='data_')
            self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        astr = '&'.join(args)
        url = '/api/instances/tagged/{}/latest?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to load tagged instance')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.TAGGED_INSTANCES)


    def get_relative(self, data_offset, data_max_length=None, tag=None, issue_date_from=None, issue_date_to=None,
                     issue_dates=None, issue_weekdays=None, issue_days=None, issue_months=None, issue_times=None,
                     data_from=None, data_to=None, time_zone=None, filter=None, function=None,
                     frequency=None, output_time_zone=None):
        """ Get a relative forecast from the TAGGED INSTANCE curve

        A relative forecast is a time series created by joining multiple
        instances so that the data date is issue_date + data_offset.
        If the data frequency is higher than the frequency of the selected
        instances, a range of data values from each instance will be used.
        Similarly if the issue frequency is higher than the data frequency,
        the same data date will be used from several instances.

        Parameters allow control of the set of instances used and how they
        are joined together, as well as post-processing of the result.

        Parameters
        ----------

        data_offset: duration, mandatory
            The duration added to the issue_date to find the start of the data
            fragment.  Format is an ISO-8601 duration string up to "days".

        data_max_length: duration, optional
            The longest duration selected from a single instance, mostly used
            to detect missing instances.  ISO-8601 duration string up to "days".

        tag: str, optional
            tag to get the data for. If omitted, the default tag is used.

        issue_date_from: time-stamp, optional
            Limits the timerange used to select instances.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_date_to: time-stamp, optional
            Limits the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "issue_date_from".

        issue_dates: list of time-stamps, optional
            List of timestamps used to select instances.
            The time-stamps can be provided in the same types as
            "issue_date_from".

        issue_weekdays: list of strings, optional
            Limits the instances to those matching the given weekdays.

        issue_days: list of integers, optional
            Limits the instances to those matching the given days of month.

        issue_months: list of strings, optional
            Limits the instances to those matching the given months.

        issue_times: list of strings, optional
            Limits the instances to those matching the given times of day.
            Format is 'HH', 'HH:mm' or 'HH:mm:ss'.

        data_from: time-stamp, optional
            start date (and time) of data to be fetched. If not given, the start
            date of the returned timeseries will be determined by the first
            selected instance and the data_offset parameter. If only the date
            (without time) is given, the time is assumed to be 00:00. The
            timestamp can be provided in the same types as "issue_date_from".

        data_to: time-stamp, optional
            end date (and time) of data to be fetched. The time-stamp can be
            provided in the same types as "issue_date_from".
            End dates are always excluded in the result!
            If not given, the end date of the returned timeseries will be
            the last date with data available in the last instance selected,
            optionally limited by the data_max_length parameter.

        time_zone: str, optional
            Change curve time zone BEFORE performing an aggregation/split
            or applying a filter. If no aggregation/split or filter is applied,
            this will simply change the timezone of the curve. Note that if
            "output_time_zone" is given, this will define the timezone of the
            returned curve, since it is applied AFTER performing an
            aggregation/split or applying a filter and thus AFTER changing
            to given "time_zone" here.

            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_time_zones`.

        filter: str, optional
            only get a specific subset of the data.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_filters` :

        function: str, optional
            function used to aggregate or split data, must be used together
            with the ``frequency`` parameter.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_functions` :

        frequency: str, optional
            data will be aggregated or split to the requested frequency using
            the given function.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        output_time_zone: str, optional
            Change curve time zone AFTER performing an aggregation/split
            or applying a filter.

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        args = [util.make_arg('data_offset', '{}'.format(data_offset))]
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        self._add_from_to(args, data_from, data_to, prefix='data_')
        self._add_functions(args, time_zone, filter, function, frequency, output_time_zone)
        if data_max_length is not None:
            args.append(util.make_arg('data_max_length', data_max_length))
        if tag is not None:
            args.append(util.make_arg('tag', tag))
        if issue_dates is not None:
            args.append(util.make_arg('issue_date', issue_dates))
        if issue_weekdays is not None:
            args.append(util.make_arg('issue_weekday', issue_weekdays))
        if issue_days is not None:
            args.append(util.make_arg('issue_day', issue_days))
        if issue_months is not None:
            args.append(util.make_arg('issue_month', issue_months))
        if issue_times is not None:
            args.append(util.make_arg('issue_time', issue_times))
        astr = '&'.join(args)
        url = '/api/instances/tagged/{}/relative?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find instances')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.TAGGED_INSTANCES)

    def get_absolute(self, data_date, issue_frequency=None, tag=None, issue_date_from=None, issue_date_to=None):
        """ Get an absolute forecast from the INSTANCE curve

        An absolute forecast is a time series created by selecting a single
        data date from a range of instances, using the issue_date as
        data date in the result.

        Parameters
        ----------

        data_date: time-stamp, mandatory
            The data date of the absolute forecast.
            The time-stamp can be provided in any of the following types :

            * datestring in format '%Y-%M-%DT%h:%m:%sZ',
              eg '2017-01-01' or '2018-12-16T13:45:00Z'
            * pandas.Timestamp object
            * datetime.datetime object

        issue_frequency: str, optional
            The frequency of the returned time series.  Mandatory if the
            issue frequency of the curve is 'ANY'.  Data will only be
            returned for those instances whose issue_date is compatible
            with issue_frequency.
            You can find valid values for this by calling
            :meth:`wapi.session.Session.get_frequencies`.

        tag: str, optional
            tag to get the data for. If omitted, the default tag is used.

        issue_date_from: time-stamp, optional
            The start of the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "data_date".

        issue_date_to: time-stamp, optional
            The end of the timerange used to select instances.
            The time-stamp can be provided in the same types as
            "data_date".

        Returns
        -------
        :class:`wapi.util.TS` object
        """
        args = [util.make_arg('data_date', data_date)]
        if issue_frequency is not None:
            args.append(util.make_arg('issue_frequency', issue_frequency))
        if tag is not None:
            args.append(util.make_arg('tag', tag))
        self._add_from_to(args, issue_date_from, issue_date_to, prefix='issue_date_')
        astr = '&'.join(args)
        url = '/api/instances/tagged/{}/absolute?{}'.format(self.id, astr)
        result = self._load_data(url, 'Failed to find instances')
        if result is None:
            return result
        return util.TS(input_dict=result, curve_type=util.TAGGED_INSTANCES)



#
# Various utility and conversion functions to make it easier to work with
# the data from the backend
#

import calendar
import datetime
import dateutil.parser
import pytz
import pandas as pd
import numpy as np
from past.types import basestring
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus


# Curve types
TIME_SERIES = 'TIME_SERIES'
TAGGED = 'TAGGED'
INSTANCES = 'INSTANCES'
TAGGED_INSTANCES = 'TAGGED_INSTANCES'


# Frequency mapping from TS to Pandas
_TS_FREQ_TABLE = {
    'Y': 'AS',
    'S': '2QS',
    'Q': 'QS',
    'M': 'MS',
    'W': 'W-MON',
    'H12': '12H',
    'H6': '6H',
    'H3': '3H',
    'MIN30': '30T',
    'MIN15': '15T',
    'MIN5': '5T',
    'MIN': 'T',
}
# Mapping from Pandas to TS is built from map above, with some additions
_PANDAS_FREQ_TABLE = {
    'AS-JAN': 'Y',
    'YS': 'Y',
    '2QS-JAN': 'S',
    'QS-JAN': 'Q',
    'min': 'MIN',
}
for k, v in _TS_FREQ_TABLE.items():
    _PANDAS_FREQ_TABLE[v] = k


class CurveException(Exception):
    pass


class TS(object):
    """
    A class to hold a basic time series.
    """
    def __init__(self, id=None, name=None, frequency=None, time_zone=None, tag=None, issue_date=None,
                 curve_type=None, points=None, input_dict=None):
        self.id = id
        self.name = name
        self.frequency = frequency
        self.time_zone = time_zone
        self.tag = tag
        self.issue_date = issue_date
        self.curve_type = curve_type
        self.points = points
        #
        # input_dict is the json dict from WAPI
        if input_dict is not None:
            for k, v in input_dict.items():
                setattr(self, k, v)

        if self.time_zone is not None:
            self.tz = parse_tz(self.time_zone)
        else:
            self.tz = pytz.timezone('CET')

        if self.curve_type is None:
            self.curve_type = detect_curve_type(self.issue_date, self.tag)
        # Validation
        if self.frequency is None:
            raise CurveException('TS must have frequency')

    def __str__(self):
        size = ''
        if self.points:
            size = ' size: {}'.format(len(self.points))
        return 'TS: {}{}'.format(self.fullname, size)

    @property
    def fullname(self):
        attrs = []
        if self.name:
            attrs.append(self.name)
        else:
            if self.id:
                attrs.append(str(self.id))
            attrs.extend([self.curve_type, self.tz.zone, self.frequency])
        if self.tag:
            attrs.append(self.tag)
        if self.issue_date:
            attrs.append(str(self.issue_date))
        return ' '.join(attrs)

    def to_pandas(self, name=None):
        """ Converting :class:`wapi.util.TS` object to a pandas.Series object

        Parameters
        ----------
        name: str, optional
            Name of the returned pandas.Series object. If not given the name
            of the curve will be used.
        Returns
        -------
        pandas.Series
        """
        if name is None:
            name = self.fullname
        if self.points is None or len(self.points) == 0:
            return pd.Series(name=name)

        index = []
        values = []
        for row in self.points:
            if len(row) != 2:
                raise ValueError('Points have unexpected contents')
            dt = datetime.datetime.fromtimestamp(row[0] / 1000.0, self.tz)
            index.append(dt)
            values.append(row[1])
        res = pd.Series(name=name, index=index, data=values)
        return res.asfreq(self._map_freq(self.frequency))

    @staticmethod
    def from_pandas(pd_series):
        # Clean up some of the more common Pandas/Wapi problems
        pd_series = pd_series.astype(np.float64)
        pd_series.replace({np.nan: None}, inplace=True)

        name = pd_series.name
        frequency = TS._rev_map_freq(pd_series.index.freqstr)

        points = []
        for i in pd_series.index:
            t = i.astimezone(pytz.utc)
            timestamp = int(calendar.timegm(t.timetuple()) * 1000)
            points.append([timestamp, pd_series[i]])

        if is_integer(name):
            return TS(id=int(name), frequency=frequency, points=points)
        else:
            return TS(name=name, frequency=frequency, points=points)

    @staticmethod
    def _map_freq(frequency):
        if frequency.upper() in _TS_FREQ_TABLE:
            frequency = _TS_FREQ_TABLE[frequency.upper()]
        return frequency

    @staticmethod
    def _rev_map_freq(frequency):
        if frequency.upper() in _PANDAS_FREQ_TABLE:
            frequency = _PANDAS_FREQ_TABLE[frequency.upper()]
        return frequency

    @staticmethod
    def sum(ts_list, name):
        """ calculate the sum of a given list of :class:`wapi.util.TS` objects

        Returns a :class:`~wapi.util.TS` (:class:`wapi.util.TS`) object that is
        the sum of a list of
        TS objects with the given name.

        Parameters
        ----------
        ts_list: list
            list of TS objects
        name: str
            Name of the returned TS object.
        Returns
        -------
        :class:`wapi.util.TS` object
        """
        df = _ts_list_to_dataframe(ts_list)
        return _generated_series_to_TS(df.sum(axis=1), name)

    @staticmethod
    def mean(ts_list, name):
        """ calculate the mean of a given list of TS objects

        Returns a TS (:class:`wapi.util.TS`) object that is
        the mean of a list of
        TS objects with the given name.

        Parameters
        ----------
        ts_list: list
            list of TS objects
        name: str
            Name of the returned TS object.
        Returns
        -------
        :class:`wapi.util.TS` object
        """
        df = _ts_list_to_dataframe(ts_list)
        return _generated_series_to_TS(df.mean(axis=1), name)

    @staticmethod
    def median(ts_list, name):
        """ calculate the median of a given list of TS objects

        Returns a TS (:class:`wapi.util.TS`) object that is
        the median of a list of
        TS objects with the given name.

        Parameters
        ----------
        ts_list: list
            list of TS objects
        name: str
            Name of the returned TS object.
        Returns
        -------
        :class:`wapi.util.TS` object
        """
        df = _ts_list_to_dataframe(ts_list)
        return _generated_series_to_TS(df.median(axis=1), name)


def _generated_series_to_TS(series, name):
    series.name = name
    return TS.from_pandas(series)


def _ts_list_to_dataframe(ts_list):
    pd_list = []
    for ts in ts_list:
        pd_list.append(ts.to_pandas())

    return pd.concat(pd_list, axis=1)


def tags_to_DF(tagged_list):
    """
    Given a list of tagged series/instances, create a DataFrame with the tag of each as column name
    """
    return pd.DataFrame({s.tag: s.to_pandas() for s in tagged_list})


#
# Some parsing helpers
#


def parsetime(datestr, tz=None):
    """
    Parse the input date and optionally convert to correct time zone
    """

    d = dateutil.parser.parse(datestr)

    if tz is not None:
        if not isinstance(tz, datetime.tzinfo):
            tz = parse_tz(tz)

        if d.tzinfo is not None:
            d = d.astimezone(tz)
        else:
            d = tz.localize(d)

    else:
        # If datestr does not have tzinfo and no tz given, assume CET
        if d.tzinfo is None:
            d = pytz.timezone('CET').localize(d)
    return d


def parserange(rangeobj, tz=None):
    """
    Parse a range object (a pair of date strings, which may each be None)
    """
    if rangeobj.get('empty') is True:
        return None
    begin = rangeobj.get('begin')
    end = rangeobj.get('end')
    if begin is not None:
        begin = parsetime(begin, tz=tz)
    if end is not None:
        end = parsetime(end, tz=tz)
    return (begin, end)


_tzmap = {
    'CEGT': 'CET',
    'WEGT': 'WET',
    'PST': 'US/Pacific',
    'TRT': 'Turkey',
    'MSK': 'Europe/Moscow',
    'ART': 'America/Argentina/Buenos_Aires',
}


def parse_tz(time_zone):
    try:
        if time_zone in _tzmap:
            time_zone = _tzmap[time_zone]
        return pytz.timezone(time_zone)
    except pytz.exceptions.UnknownTimeZoneError:
        return pytz.timezone('CET')


def detect_curve_type(issue_date, tag):
    if issue_date is None and tag is None:
        return TIME_SERIES
    elif issue_date is None:
        return TAGGED
    elif tag is None:
        return INSTANCES
    else:
        return TAGGED_INSTANCES


def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def make_arg(key, value):
    if hasattr(value, '__iter__') and not isinstance(value, basestring):
        return '&'.join([make_arg(key, v) for v in value])

    if isinstance(value, datetime.date):
        tmp = value.isoformat()
    else:
        tmp = '{}'.format(value)
    v = quote_plus(tmp)
    return '{}={}'.format(key, v)



import json
import time

import sseclient
import threading
import queue

from . import curves, util
from builtins import str


class EventListener:
    def __init__(self, session, curve_list, start_time=None, timeout=None):
        self.curve_cache = {}
        ids = []
        if not hasattr(curve_list, '__iter__') or isinstance(curve_list, str):
            curve_list = [curve_list]
        for curve in curve_list:
            if isinstance(curve, curves.BaseCurve):
                ids.append(curve.id)
                self.curve_cache[curve.id] = curve
            else:
                ids.append(curve)
        args = [util.make_arg('id', ids)]
        if start_time is not None:
            args.append(util.make_arg('start_time', start_time))
        self.url = '/api/events?{}'.format('&'.join(args))
        self.session = session
        self.timeout = timeout
        self.retry = 3000 # Retry time in milliseconds
        self.client = None
        self.queue = queue.Queue()
        self.do_shutdown = False
        self.worker = threading.Thread(target=self.fetch_events)
        self.worker.setDaemon(True)
        self.worker.start()

    def get(self):
        try:
            val = self.queue.get(timeout=self.timeout)
            if isinstance(val, EventError):
                raise val.exception
            return val
        except queue.Empty:
            return EventTimeout()

    def fetch_events(self):
        while not self.do_shutdown:
            try:
                with self.session.data_request("GET", self.session.urlbase, self.url, stream=True) as stream:
                    self.client = sseclient.SSEClient(stream)
                    for sse_event in self.client.events():
                        if sse_event.event == 'curve_event':
                            event = CurveEvent(sse_event)
                        else:
                            event = DefaultEvent(sse_event)
                        if hasattr(event, 'id') and event.id in self.curve_cache:
                            event.curve = self.curve_cache[event.id]
                        self.queue.put(event)
                        if sse_event.retry is not None:
                            try:
                                self.retry = int(sse_event.retry)
                            except:
                                pass
                        if self.do_shutdown:
                            break
                    # Session was closed by server/network, wait for retry before looping.
                    time.sleep(self.retry / 1000.0)
            except Exception as e:
                self.queue.put(EventError(e))
                break

    def close(self, timeout=1):
        self.do_shutdown = True
        if self.client is not None:
            self.client.close()
        self.worker.join(timeout)

    def __iter__(self):
        return self

    def __next__(self):
        return self.get()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class EventError:
    def __init__(self, exception):
        self.exception = exception

    def __str__(self):
        return "{}".format(self.exception)


class EventTimeout:
    """Returned on timeout, etc."""
    pass


class DefaultEvent(object):
    def __init__(self, sse_event):
        self._raw_event = sse_event
        try:
            self.json_data = json.loads(sse_event.data)
        except:
            self.json_data = None


class CurveEvent(DefaultEvent):
    def __init__(self, sse_event):
        super(CurveEvent, self).__init__(sse_event)
        self.id = self.json_data['id']
        self.curve = None
        self.created = util.parsetime(self.json_data['created'])
        self.operation = self.json_data['operation']
        self.tag = None
        self.issue_date = None
        self.range = None
        if 'tag' in self.json_data:
            self.tag = self.json_data['tag']
        if 'issue_date' in self.json_data:
            self.issue_date = util.parsetime(self.json_data['issue_date'])
        if 'range' in self.json_data:
            self.range = util.parserange(self.json_data['range'])



# encoding: utf-8
#
import os, sys
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
# Get current version from the VERSION file
with open(os.path.join(here, 'VERSION')) as fv:
    version = fv.read()

install_requires = [
    'requests >= 2.18',
    'sseclient-py >= 1.7',
    'pytz',
    'pandas >= 0.21',
    'future >= 0.16',
]
if sys.version_info < (3,):
    install_requires.append('configparser >= 3.5')

setup(
    name='wapi-python',
    packages=['wapi'],
    install_requires=install_requires,
    tests_require=[
        'pytest',
        'pytest-cov >= 2.5',
        'requests-mock >= 1.3',
    ],
    version=version,
    description='Wattsight API python library',
    long_description='This library is meant as a simple toolkit for working with data from https://api.wattsight.com/ (or equivalent services).  Note that access is based on some sort of login credentials, this library is not all that useful unless you have a valid Wattsight account.',
    author='Wattsight',
    author_email='support@wattsight.com',
    url='https://www.wattsight.com'
)



try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

import requests
import json
import time
import warnings
from past.types import basestring
import configparser

from . import auth, curves, events, util
from .util import CurveException


RETRY_COUNT = 4    # Number of times to retry
RETRY_DELAY = 0.5  # Delay between retried calls, in seconds.
TIMEOUT = 300      # Default timeout for web calls, in seconds.
API_URLBASE = 'https://api.wattsight.com'
AUTH_URLBASE = 'https://auth.wattsight.com'


class ConfigException(Exception):
    pass


class MetadataException(Exception):
    pass


class Session(object):
    """ Establish a connection to Wattsight API

    Creates an object that holds the state which is needed when talking to the
    Wattsight data center. To establish a session, you have to provide
    suthentication information either directly by using a ```client_id` and
    ``client_secret`` or using a ``config_file`` .

    See https://api.wattsight.com/#documentation for information how to get
    your authentication data.

    Parameters
    ----------

    urlbase: url
        Location of Wattsight service
    config_file: path
        path to the config.ini file which contains your authentication
        information.
    client_id: str
        Your client ID
    client_secret:
        Your client secret.
    auth_urlbase: url
        Location of Wattsight authentication service
    timeout: float
        Timeout for REST calls, in seconds

    Returns
    -------
    session: :class:`wapi.session.Session` object

    """

    def __init__(self, urlbase=None, config_file=None, client_id=None, client_secret=None,
                 auth_urlbase=None, timeout=None):
        self.urlbase = API_URLBASE
        self.auth = None
        self.timeout = TIMEOUT
        self._session = requests.Session()
        if config_file is not None:
            self.read_config_file(config_file)
        elif client_id is not None and client_secret is not None:
            self.configure(client_id, client_secret, auth_urlbase)
        if urlbase is not None:
            self.urlbase = urlbase
        if timeout is not None:
            self.timeout = timeout

    def read_config_file(self, config_file):
        """Set up according to configuration file with hosts and access details"""
        if self.auth is not None:
            raise ConfigException('Session configuration is already done')
        config = configparser.RawConfigParser()
        # Support being given a file-like object or a file path:
        if hasattr(config_file, 'read'):
            config.read_file(config_file)
        else:
            files_read = config.read(config_file)
            if not files_read:
                raise ConfigException('Configuration file with name {} '
                                      'was not found.'.format(config_file))
        urlbase = config.get('common', 'urlbase', fallback=None)
        if urlbase is not None:
            self.urlbase = urlbase
        auth_type = config.get('common', 'auth_type')
        if auth_type == 'OAuth':
            client_id = config.get(auth_type, 'id')
            client_secret = config.get(auth_type, 'secret')
            auth_urlbase = config.get(auth_type, 'auth_urlbase', fallback=AUTH_URLBASE)
            self.auth = auth.OAuth(self, client_id, client_secret, auth_urlbase)
        timeout = config.get('common', 'timeout', fallback=None)
        if timeout is not None:
            self.timeout = float(timeout)

    def configure(self, client_id, client_secret, auth_urlbase=None):
        """Programmatically set authentication parameters"""
        if self.auth is not None:
            raise ConfigException('Session configuration is already done')
        if auth_urlbase is None:
            auth_urlbase = AUTH_URLBASE
        self.auth = auth.OAuth(self, client_id, client_secret, auth_urlbase)

    def get_curve(self, id=None, name=None):
        """Getting a curve object

        Return a curve object of the correct type.  Name should be specified.
        While it is possible to get a curve by id, this is not guaranteed to be
        long-term stable and will be removed in future versions.

        Parameters
        ----------

        id: int
            curve id (deprecated)
        name: str
            curve name

        Returns
        -------
        curve object
            Curve objects, can be one of:
            :class:`~wapi.curves.TimeSeriesCurve`,
            :class:`~wapi.curves.TaggedCurve`,
            :class:`~wapi.curves.InstanceCurve`,
            :class:`~wapi.curves.TaggedInstanceCurve`.
        """
        if id is not None:
            warnings.warn("Looking up a curve by ID will be removed in the future.", FutureWarning, stacklevel=2)
        if id is None and name is None:
            raise MetadataException('No curve specified')

        if id is not None:
            arg = util.make_arg('id', id)
        else:
            arg = util.make_arg('name', name)
        response = self.data_request('GET', self.urlbase, '/api/curves/get?{}'.format(arg))
        return self.handle_single_curve_response(response)

    def search(self, query=None, id=None, name=None, commodity=None, category=None, area=None, station=None,
               source=None, scenario=None, unit=None, time_zone=None, version=None, frequency=None, data_type=None,
               curve_state=None, modified_since=None, only_accessible=None):
        """
        Search for a curve matching various metadata.

        This function searches for curves that matches the given search
        parameters and returns a list of 0 or more curve objects.
        A curve object can be a
        :class:`~wapi.curves.TimeSeriesCurve`,
        :class:`~wapi.curves.TaggedCurve`,
        :class:`~wapi.curves.InstanceCurve` or a
        :class:`~wapi.curves.TaggedInstanceCurve` object.

        The search will return those curves matching all supplied parameters
        (logical AND). For most parameters, a list of values may be supplied.
        The search will match any of these values (logical OR).  If a single
        value contains a string with comma-separated values, these will be
        treated as a list but will match with logical AND. (This only makes
        sense for parameters where a curve may have multiple values:
        area (border curves), category, source and scenario.)

        For more details, see the REST documentation.

        Parameters
        ----------

        query: str
            A query string used for a language-aware text search on both names
            and descriptions of the various attributes in the curve.

        id: int or lits of int
            search for one or more specific id's (deprecated)

        name: str or list of str
            search for one or more curve names, you can use the ``*`` as
            a wildcard for patter matching.

        commodity: str or list of str
            search for curves that match the given ``commodity`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_commodities`

        category: str or list of str
            search for curves that match the given ``category`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_categories`

        area: str or list of str
            search for curves that match the given ``area`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_areas`

        station: str or list of str
            search for curves that match the given ``station`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_stations`

        source: str or list of str
            search for curves that match the given ``source`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_sources`

        scenario: str or list of str
            search for curves that match the given ``scenario`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_scenarios`

        unit: str or list of str
            search for curves that match the given ``unit`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_units`

        time_zone: str or list of str
            search for curves that match the given ``time_zone`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_time_zones`

        version: str or list of str
            search for curves that match the given ``version`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_versions`

        frequency: str or list of str
            search for curves that match the given ``frequency`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_frequencies`

        data_type: str or list of str
            search for curves that match the given ``data_type`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_data_types`

        curve_state: str or list of str
            search for curves that match the given ``curve_state`` attribute.
            Get valid values for this attribute with
            :meth:`wapi.session.Session.get_curve_state`

        modified_since: datestring, pandas.Timestamp or datetime.datetime
            only return curves that where modified after given datetime.

        only_accessible: bool
            If True, only return curves you have (some) access to.

        Returns
        -------
        curves: list
            list of curve objects, can be one of:
            :class:`~wapi.curves.TimeSeriesCurve`,
            :class:`~wapi.curves.TaggedCurve`,
            :class:`~wapi.curves.InstanceCurve`,
            :class:`~wapi.curves.TaggedInstanceCurve`.
        """
        search_terms = {
            'query': query,
            'id': id,
            'name': name,
            'commodity': commodity,
            'category': category,
            'area': area,
            'station': station,
            'source': source,
            'scenario': scenario,
            'unit': unit,
            'time_zone': time_zone,
            'version': version,
            'frequency': frequency,
            'data_type': data_type,
            'curve_state': curve_state,
            'modified_since': modified_since,
            'only_accessible': only_accessible,
        }
        if id is not None:
            warnings.warn("Searching for curves by ID will be removed in the future.", FutureWarning, stacklevel=2)
        args = []
        astr = ''
        for key, val in search_terms.items():
            if val is None:
                continue
            args.append(util.make_arg(key, val))
        if len(args):
            astr = "?{}".format("&".join(args))
        # Now run the search, and try to produce a list of curves
        response = self.data_request('GET', self.urlbase, '/api/curves{}'.format(astr))
        return self.handle_multi_curve_response(response)

    def make_curve(self, id, curve_type):
        """Return a mostly uninitialized curve object of the correct type.
        This is generally a bad idea, use get_curve or search when possible."""
        if curve_type in self._curve_types:
            return self._curve_types[curve_type](id, None, self)
        raise CurveException('Bad curve type requested')

    def events(self, curve_list, start_time=None, timeout=None):
        """Get an event listener for a list of curves."""
        return events.EventListener(self, curve_list, start_time=start_time, timeout=timeout)

    _attributes = {'commodities', 'categories', 'areas', 'stations', 'sources', 'scenarios',
                   'units', 'time_zones', 'versions', 'frequencies', 'data_types',
                   'curve_states', 'curve_types', 'functions', 'filters'}

    def get_commodities(self):
        """
        Get valid values for the commodity attribute
        """
        return self.get_attribute('commodities')

    def get_categories(self):
        """
        Get valid values for the category attribute
        """
        return self.get_attribute('categories')

    def get_areas(self):
        """
        Get valid values for the area attribute
        """
        return self.get_attribute('areas')

    def get_stations(self):
        """
        Get valid values for the station attribute
        """
        return self.get_attribute('stations')

    def get_sources(self):
        """
        Get valid values for the source attribute
        """
        return self.get_attribute('sources')

    def get_scenarios(self):
        """
        Get valid values for the scenarios attribute
        """
        return self.get_attribute('scenarios')

    def get_units(self):
        """
        Get valid values for the unit attribute
        """
        return self.get_attribute('units')

    def get_time_zones(self):
        """
        Get valid values for the time zone attribute
        """
        return self.get_attribute('time_zones')

    def get_versions(self):
        """
        Get valid values for the version attribute
        """
        return self.get_attribute('versions')

    def get_frequencies(self):
        """
        Get valid values for the frequency attribute
        """
        return self.get_attribute('frequencies')

    def get_data_types(self):
        """
        Get valid values for the data_type attribute
        """
        return self.get_attribute('data_types')

    def get_curve_states(self):
        """
        Get valid values for the curve_state attribute
        """
        return self.get_attribute('curve_states')

    def get_curve_types(self):
        """
        Get valid values for the curve_type attribute
        """
        return self.get_attribute('curve_types')

    def get_functions(self):
        """
        Get valid values for the function attribute
        """
        return self.get_attribute('functions')

    def get_filters(self):
        """
        Get valid values for the filter attribute
        """
        return self.get_attribute('filters')

    def get_attribute(self, attribute):
        """Get valid values for an attribute."""
        if attribute not in self._attributes:
            raise MetadataException('Attribute {} is not valid'.format(attribute))
        response = self.data_request('GET', self.urlbase, '/api/{}'.format(attribute))
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 204:
            return None
        raise MetadataException('Failed loading {}: {}'.format(attribute,
                                                               response.content.decode()))

    _curve_types = {
        util.TIME_SERIES:      curves.TimeSeriesCurve,
        util.TAGGED:           curves.TaggedCurve,
        util.INSTANCES:        curves.InstanceCurve,
        util.TAGGED_INSTANCES: curves.TaggedInstanceCurve,
    }

    _meta_keys = ('id', 'name', 'frequency', 'time_zone', 'curve_type')

    def _build_curve(self, metadata):
        for key in self._meta_keys:
            if key not in metadata:
                raise MetadataException('Mandatory key {} not found in metadata'.format(key))
        curve_id = int(metadata['id'])
        if metadata['curve_type'] in self._curve_types:
            c = self._curve_types[metadata['curve_type']](curve_id, metadata, self)
            return c
        raise CurveException('Unknown curve type ({})'.format(metadata['curve_type']))

    def data_request(self, req_type, urlbase, url, data=None, rawdata=None, authval=None,
                     stream=False, retries=RETRY_COUNT):
        """Run a call to the backend, dealing with authentication etc."""
        headers = {}

        if not urlbase:
            urlbase = self.urlbase
        longurl = urljoin(urlbase, url)

        databytes = None
        if data is not None:
            headers['content-type'] = 'application/json'
            if isinstance(data, basestring):
                databytes = data.encode()
            else:
                databytes = json.dumps(data).encode()
        if data is None and rawdata is not None:
            databytes = rawdata
        if self.auth is not None:
            self.auth.validate_auth()
            headers.update(self.auth.get_headers(databytes))
        timeout = None
        try:
            res = self._session.request(method=req_type, url=longurl, data=databytes,
                                        headers=headers, auth=authval, stream=stream, timeout=self.timeout)
        except requests.exceptions.Timeout as e:
            timeout = e
            res = None
        if (timeout is not None or (500 <= res.status_code < 600) or res.status_code == 408) and retries > 0:
            if RETRY_DELAY > 0:
                time.sleep(RETRY_DELAY)
            return self.data_request(req_type, urlbase, url, data, rawdata, authval, stream, retries-1)
        if timeout is not None:
            raise timeout
        return res

    def handle_single_curve_response(self, response):
        if not response.ok:
            raise MetadataException('Failed to load curve: {}'
                                    .format(response.content.decode()))
        metadata = response.json()
        return self._build_curve(metadata)

    def handle_multi_curve_response(self, response):
        if not response.ok:
            raise MetadataException('Curve search failed: {}'
                                    .format(response.content.decode()))
        metadata_list = response.json()

        result = []
        for metadata in metadata_list:
            result.append(self._build_curve(metadata))
        return result



#
# Authentication support
#

import json
import time

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


class AuthFailedException(Exception):
    pass


class OAuth:
    """
    Authentication based on OAuth client ID.
    This is the main authentication mechanism for customer access to the data center.
    """

    def __init__(self, session, client_id, client_secret, auth_urlbase):
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_urlbase = auth_urlbase
        self.token = None
        self.token_type = None
        self.valid_until = None
        self.session = session
        self.revalidating = False
        self._authenticate()

    def validate_auth(self):
        """Check valid_until and fetch new token if needed"""
        # Prevent loop
        if self.revalidating:
            return
        if self.valid_until is None or time.time() > self.valid_until:
            self.revalidating = True
            self._authenticate()
            self.revalidating = False

    def _authenticate(self):
        # Wipe out any old values before (re-)login
        self.token = None
        self.token_type = None
        self.valid_until = None
        now = time.time()
        url = urljoin(self.auth_urlbase, '/oauth2/token')
        auth = (self.client_id, self.client_secret)
        data = {'grant_type': 'client_credentials'}
        response = self.session.data_request('POST', self.auth_urlbase, url, rawdata=data, authval=auth)
        if response.status_code != 200:
            raise AuthFailedException('Authentication failed: {}'.format(response.content))
        # Parse token
        rsp = json.loads(response.content.decode())
        self.token = rsp['access_token']
        self.token_type = rsp['token_type']
        self.valid_until = now + int(rsp['expires_in'] * 0.95)

    def get_headers(self, data):
        """The web-token auth header is simple"""
        if self.token is not None and self.token_type is not None:
            return {'Authorization': '{} {}'.format(self.token_type, self.token)}
        return {}
